@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2005-2009 Junjiro R. Okajima
+ * Copyright (C) 2005-2010 Junjiro R. Okajima
  *
  * This program, aufs is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -28,42 +28,7 @@
 #include <linux/security.h>
 #include "aufs.h"
 
-/* common function to regular file and dir */
-int aufs_flush(struct file *file, fl_owner_t id)
-{
-	int err;
-	aufs_bindex_t bindex, bend;
-	struct dentry *dentry;
-	struct file *h_file;
-
-	dentry = file->f_dentry;
-	si_noflush_read_lock(dentry->d_sb);
-	fi_read_lock(file);
-	di_read_lock_child(dentry, AuLock_IW);
-
-	err = 0;
-	bend = au_fbend(file);
-	for (bindex = au_fbstart(file); !err && bindex <= bend; bindex++) {
-		h_file = au_h_fptr(file, bindex);
-		if (!h_file || !h_file->f_op || !h_file->f_op->flush)
-			continue;
-
-		err = h_file->f_op->flush(h_file, id);
-		if (!err)
-			vfsub_update_h_iattr(&h_file->f_path, /*did*/NULL);
-		/*ignore*/
-	}
-	au_cpup_attr_timesizes(dentry->d_inode);
-
-	di_read_unlock(dentry, AuLock_IW);
-	fi_read_unlock(file);
-	si_read_unlock(dentry->d_sb);
-	return err;
-}
-
-/* ---------------------------------------------------------------------- */
-
-static int do_open_nondir(struct file *file, int flags)
+int au_do_open_nondir(struct file *file, int flags)
 {
 	int err;
 	aufs_bindex_t bindex;
@@ -76,8 +41,8 @@ static int do_open_nondir(struct file *file, int flags)
 	err = 0;
 	dentry = file->f_dentry;
 	finfo = au_fi(file);
-	finfo->fi_h_vm_ops = NULL;
-	finfo->fi_vm_ops = NULL;
+	memset(&finfo->fi_htop, 0, sizeof(finfo->fi_htop));
+	finfo->fi_hvmop = NULL;
 	bindex = au_dbstart(dentry);
 	/* O_TRUNC is processed already */
 	BUG_ON(au_test_ro(dentry->d_sb, bindex, dentry->d_inode)
@@ -88,7 +53,6 @@ static int do_open_nondir(struct file *file, int flags)
 		err = PTR_ERR(h_file);
 	else {
 		au_set_fbstart(file, bindex);
-		au_set_fbend(file, bindex);
 		au_set_h_fptr(file, bindex, h_file);
 		au_update_figen(file);
 		/* todo: necessary? */
@@ -100,23 +64,51 @@ static int do_open_nondir(struct file *file, int flags)
 static int aufs_open_nondir(struct inode *inode __maybe_unused,
 			    struct file *file)
 {
-	return au_do_open(file, do_open_nondir);
+	int err;
+	struct super_block *sb;
+
+	sb = file->f_dentry->d_sb;
+	si_read_lock(sb, AuLock_FLUSH);
+	err = au_do_open(file, au_do_open_nondir, /*fidir*/NULL);
+	si_read_unlock(sb);
+	return err;
 }
 
-static int aufs_release_nondir(struct inode *inode __maybe_unused,
-			       struct file *file)
+int aufs_release_nondir(struct inode *inode __maybe_unused, struct file *file)
 {
-	struct super_block *sb = file->f_dentry->d_sb;
+	struct au_finfo *finfo;
+	aufs_bindex_t bindex;
+
+	finfo = au_fi(file);
+	bindex = finfo->fi_btop;
+	if (bindex >= 0)
+		au_set_h_fptr(file, bindex, NULL);
 
-	si_noflush_read_lock(sb);
-	kfree(au_fi(file)->fi_vm_ops);
 	au_finfo_fin(file);
-	si_read_unlock(sb);
 	return 0;
 }
 
 /* ---------------------------------------------------------------------- */
 
+static int au_do_flush_nondir(struct file *file, fl_owner_t id)
+{
+	int err;
+	struct file *h_file;
+
+	err = 0;
+	h_file = au_hf_top(file);
+	if (h_file)
+		err = vfsub_flush(h_file, id);
+	return err;
+}
+
+static int aufs_flush_nondir(struct file *file, fl_owner_t id)
+{
+	return au_do_flush(file, id, au_do_flush_nondir);
+}
+
+/* ---------------------------------------------------------------------- */
+
 static ssize_t aufs_read(struct file *file, char __user *buf, size_t count,
 			 loff_t *ppos)
 {
@@ -132,7 +124,7 @@ static ssize_t aufs_read(struct file *file, char __user *buf, size_t count,
 	if (unlikely(err))
 		goto out;
 
-	h_file = au_h_fptr(file, au_fbstart(file));
+	h_file = au_hf_top(file);
 	err = vfsub_read_u(h_file, buf, count, ppos);
 	/* todo: necessary? */
 	/* file->f_ra = h_file->f_ra; */
@@ -149,7 +141,6 @@ static ssize_t aufs_write(struct file *file, const char __user *ubuf,
 			  size_t count, loff_t *ppos)
 {
 	ssize_t err;
-	aufs_bindex_t bstart;
 	struct au_pin pin;
 	struct dentry *dentry;
 	struct inode *inode;
@@ -172,8 +163,7 @@ static ssize_t aufs_write(struct file *file, const char __user *ubuf,
 	if (unlikely(err))
 		goto out_unlock;
 
-	bstart = au_fbstart(file);
-	h_file = au_h_fptr(file, bstart);
+	h_file = au_hf_top(file);
 	au_unpin(&pin);
 	err = vfsub_write_u(h_file, buf, count, ppos);
 	au_cpup_attr_timesizes(inode);
@@ -188,6 +178,37 @@ static ssize_t aufs_write(struct file *file, const char __user *ubuf,
 	return err;
 }
 
+static ssize_t au_do_aio(struct file *h_file, int rw, struct kiocb *kio,
+			 const struct iovec *iov, unsigned long nv, loff_t pos)
+{
+	ssize_t err;
+	struct file *file;
+	ssize_t (*func)(struct kiocb *, const struct iovec *, unsigned long,
+			loff_t);
+
+	err = security_file_permission(h_file, rw);
+	if (unlikely(err))
+		goto out;
+
+	err = -ENOSYS;
+	func = NULL;
+	if (rw == MAY_READ)
+		func = h_file->f_op->aio_read;
+	else if (rw == MAY_WRITE)
+		func = h_file->f_op->aio_write;
+	if (func) {
+		file = kio->ki_filp;
+		kio->ki_filp = h_file;
+		err = func(kio, iov, nv, pos);
+		kio->ki_filp = file;
+	} else
+		/* currently there is no such fs */
+		WARN_ON_ONCE(1);
+
+ out:
+	return err;
+}
+
 static ssize_t aufs_aio_read(struct kiocb *kio, const struct iovec *iov,
 			     unsigned long nv, loff_t pos)
 {
@@ -204,29 +225,14 @@ static ssize_t aufs_aio_read(struct kiocb *kio, const struct iovec *iov,
 	if (unlikely(err))
 		goto out;
 
-	err = -ENOSYS;
-	h_file = au_h_fptr(file, au_fbstart(file));
-	if (h_file->f_op && h_file->f_op->aio_read) {
-		err = security_file_permission(h_file, MAY_READ);
-		if (unlikely(err))
-			goto out_unlock;
-		if (!is_sync_kiocb(kio)) {
-			get_file(h_file);
-			fput(file);
-		}
-		kio->ki_filp = h_file;
-		err = h_file->f_op->aio_read(kio, iov, nv, pos);
-		/* todo: necessary? */
-		/* file->f_ra = h_file->f_ra; */
-		fsstack_copy_attr_atime(dentry->d_inode,
-					h_file->f_dentry->d_inode);
-	} else
-		/* currently there is no such fs */
-		WARN_ON_ONCE(h_file->f_op && h_file->f_op->read);
-
- out_unlock:
+	h_file = au_hf_top(file);
+	err = au_do_aio(h_file, MAY_READ, kio, iov, nv, pos);
+	/* todo: necessary? */
+	/* file->f_ra = h_file->f_ra; */
+	fsstack_copy_attr_atime(dentry->d_inode, h_file->f_dentry->d_inode);
 	di_read_unlock(dentry, AuLock_IR);
 	fi_read_unlock(file);
+
  out:
 	si_read_unlock(sb);
 	return err;
@@ -236,7 +242,6 @@ static ssize_t aufs_aio_write(struct kiocb *kio, const struct iovec *iov,
 			      unsigned long nv, loff_t pos)
 {
 	ssize_t err;
-	aufs_bindex_t bstart;
 	struct au_pin pin;
 	struct dentry *dentry;
 	struct inode *inode;
@@ -249,7 +254,6 @@ static ssize_t aufs_aio_write(struct kiocb *kio, const struct iovec *iov,
 	inode = dentry->d_inode;
 	mutex_lock(&inode->i_mutex);
 	si_read_lock(sb, AuLock_FLUSH);
-
 	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
 	if (unlikely(err))
 		goto out;
@@ -259,25 +263,11 @@ static ssize_t aufs_aio_write(struct kiocb *kio, const struct iovec *iov,
 	if (unlikely(err))
 		goto out_unlock;
 
-	err = -ENOSYS;
-	bstart = au_fbstart(file);
-	h_file = au_h_fptr(file, bstart);
 	au_unpin(&pin);
-	if (h_file->f_op && h_file->f_op->aio_write) {
-		err = security_file_permission(h_file, MAY_WRITE);
-		if (unlikely(err))
-			goto out_unlock;
-		if (!is_sync_kiocb(kio)) {
-			get_file(h_file);
-			fput(file);
-		}
-		kio->ki_filp = h_file;
-		err = h_file->f_op->aio_write(kio, iov, nv, pos);
-		au_cpup_attr_timesizes(inode);
-		inode->i_mode = h_file->f_dentry->d_inode->i_mode;
-	} else
-		/* currently there is no such fs */
-		WARN_ON_ONCE(h_file->f_op && h_file->f_op->write);
+	h_file = au_hf_top(file);
+	err = au_do_aio(h_file, MAY_WRITE, kio, iov, nv, pos);
+	au_cpup_attr_timesizes(inode);
+	inode->i_mode = h_file->f_dentry->d_inode->i_mode;
 
  out_unlock:
 	di_read_unlock(dentry, AuLock_IR);
@@ -305,7 +295,7 @@ static ssize_t aufs_splice_read(struct file *file, loff_t *ppos,
 		goto out;
 
 	err = -EINVAL;
-	h_file = au_h_fptr(file, au_fbstart(file));
+	h_file = au_hf_top(file);
 	if (au_test_loopback_kthread()) {
 		file->f_mapping = h_file->f_mapping;
 		smp_mb(); /* unnecessary? */
@@ -349,7 +339,7 @@ aufs_splice_write(struct pipe_inode_info *pipe, struct file *file, loff_t *ppos,
 	if (unlikely(err))
 		goto out_unlock;
 
-	h_file = au_h_fptr(file, au_fbstart(file));
+	h_file = au_hf_top(file);
 	au_unpin(&pin);
 	err = vfsub_splice_from(pipe, h_file, ppos, len, flags);
 	au_cpup_attr_timesizes(inode);
@@ -394,12 +384,13 @@ static int aufs_fault(struct vm_area_struct *vma, struct vm_fault *vmf)
 
 	/* do not revalidate, no si lock */
 	finfo = au_fi(file);
-	h_file = finfo->fi_hfile[0 + finfo->fi_bstart].hf_file;
-	AuDebugOn(!h_file || !finfo->fi_h_vm_ops);
+	AuDebugOn(finfo->fi_hdir);
+	h_file = finfo->fi_htop.hf_file;
+	AuDebugOn(!h_file || !finfo->fi_hvmop);
 
 	mutex_lock(&finfo->fi_vm_mtx);
 	vma->vm_file = h_file;
-	err = finfo->fi_h_vm_ops->fault(vma, vmf);
+	err = finfo->fi_hvmop->fault(vma, vmf);
 	/* todo: necessary? */
 	/* file->f_ra = h_file->f_ra; */
 	au_reset_file(vma, file);
@@ -424,12 +415,13 @@ static int aufs_page_mkwrite(struct vm_area_struct *vma, struct vm_fault *vmf)
 	wait_event(wq, (file = au_safe_file(vma)));
 
 	finfo = au_fi(file);
-	h_file = finfo->fi_hfile[0 + finfo->fi_bstart].hf_file;
-	AuDebugOn(!h_file || !finfo->fi_h_vm_ops);
+	AuDebugOn(finfo->fi_hdir);
+	h_file = finfo->fi_htop.hf_file;
+	AuDebugOn(!h_file || !finfo->fi_hvmop);
 
 	mutex_lock(&finfo->fi_vm_mtx);
 	vma->vm_file = h_file;
-	err = finfo->fi_h_vm_ops->page_mkwrite(vma, vmf);
+	err = finfo->fi_hvmop->page_mkwrite(vma, vmf);
 	au_reset_file(vma, file);
 	mutex_unlock(&finfo->fi_vm_mtx);
 	wake_up(&wq);
@@ -446,149 +438,226 @@ static void aufs_vm_close(struct vm_area_struct *vma)
 	wait_event(wq, (file = au_safe_file(vma)));
 
 	finfo = au_fi(file);
-	h_file = finfo->fi_hfile[0 + finfo->fi_bstart].hf_file;
-	AuDebugOn(!h_file || !finfo->fi_h_vm_ops);
+	AuDebugOn(finfo->fi_hdir);
+	h_file = finfo->fi_htop.hf_file;
+	AuDebugOn(!h_file || !finfo->fi_hvmop);
 
 	mutex_lock(&finfo->fi_vm_mtx);
 	vma->vm_file = h_file;
-	finfo->fi_h_vm_ops->close(vma);
+	finfo->fi_hvmop->close(vma);
 	au_reset_file(vma, file);
 	mutex_unlock(&finfo->fi_vm_mtx);
 	wake_up(&wq);
 }
 
-static struct vm_operations_struct aufs_vm_ops = {
-	/* .close and .page_mkwrite are not set by default */
+const struct vm_operations_struct aufs_vm_ops = {
+	.close		= aufs_vm_close,
 	.fault		= aufs_fault,
+	.page_mkwrite	= aufs_page_mkwrite
 };
 
 /* ---------------------------------------------------------------------- */
 
-static unsigned long au_prot_conv(unsigned long flags)
-{
-	unsigned long prot;
+/* cf. linux/include/linux/mman.h: calc_vm_prot_bits() */
+#define AuConv_VM_PROT(f, b)	_calc_vm_trans(f, VM_##b, PROT_##b)
 
-	prot = 0;
-	if (flags & VM_READ)
-		prot |= PROT_READ;
-	if (flags & VM_WRITE)
-		prot |= PROT_WRITE;
-	if (flags & VM_EXEC)
-		prot |= PROT_EXEC;
-	return prot;
+static unsigned long au_arch_prot_conv(unsigned long flags)
+{
+	/* currently ppc64 only */
+#ifdef CONFIG_PPC64
+	/* cf. linux/arch/powerpc/include/asm/mman.h */
+	AuDebugOn(arch_calc_vm_prot_bits(-1) != VM_SAO);
+	return AuConv_VM_PROT(flags, SAO);
+#else
+	AuDebugOn(arch_calc_vm_prot_bits(-1));
+	return 0;
+#endif
 }
 
-static struct vm_operations_struct *au_vm_ops(struct file *h_file,
-					      struct vm_area_struct *vma)
+static unsigned long au_prot_conv(unsigned long flags)
 {
-	struct vm_operations_struct *vm_ops;
+	return AuConv_VM_PROT(flags, READ)
+		| AuConv_VM_PROT(flags, WRITE)
+		| AuConv_VM_PROT(flags, EXEC)
+		| au_arch_prot_conv(flags);
+}
+
+/* cf. linux/include/linux/mman.h: calc_vm_flag_bits() */
+#define AuConv_VM_MAP(f, b)	_calc_vm_trans(f, VM_##b, MAP_##b)
+
+static unsigned long au_flag_conv(unsigned long flags)
+{
+	return AuConv_VM_MAP(flags, GROWSDOWN)
+		| AuConv_VM_MAP(flags, DENYWRITE)
+		| AuConv_VM_MAP(flags, EXECUTABLE)
+		| AuConv_VM_MAP(flags, LOCKED);
+}
+
+static struct vm_operations_struct *
+au_hvmop(struct file *h_file, struct vm_area_struct *vma, unsigned long *flags)
+{
+	struct vm_operations_struct *h_vmop;
+	unsigned long prot;
 	int err;
 
-	vm_ops = ERR_PTR(-ENODEV);
+	h_vmop = ERR_PTR(-ENODEV);
 	if (!h_file->f_op || !h_file->f_op->mmap)
 		goto out;
 
-	err = ima_file_mmap(h_file, au_prot_conv(vma->vm_flags));
-	vm_ops = ERR_PTR(err);
-	if (err)
-		goto out;
-
-	err = h_file->f_op->mmap(h_file, vma);
-	vm_ops = ERR_PTR(err);
+	prot = au_prot_conv(vma->vm_flags);
+	err = security_file_mmap(h_file, /*reqprot*/prot, prot,
+				 au_flag_conv(vma->vm_flags), vma->vm_start, 0);
+	h_vmop = ERR_PTR(err);
 	if (unlikely(err))
 		goto out;
 
-	vm_ops = vma->vm_ops;
+	err = ima_file_mmap(h_file, prot);
+	h_vmop = ERR_PTR(err);
+	if (unlikely(err))
+		goto out;
+
+	err = h_file->f_op->mmap(h_file, vma);
+	h_vmop = ERR_PTR(err);
+	if (unlikely(err))
+		goto out;
+
+	/* oops, it became 'const' */
+	h_vmop = (struct vm_operations_struct *)vma->vm_ops;
+	*flags = vma->vm_flags;
 	err = do_munmap(current->mm, vma->vm_start,
 			vma->vm_end - vma->vm_start);
 	if (unlikely(err)) {
 		AuIOErr("failed internal unmapping %.*s, %d\n",
 			AuDLNPair(h_file->f_dentry), err);
-		vm_ops = ERR_PTR(-EIO);
+		h_vmop = ERR_PTR(-EIO);
 	}
 
  out:
-	return vm_ops;
+	return h_vmop;
 }
 
-static int au_custom_vm_ops(struct au_finfo *finfo, struct vm_area_struct *vma)
+/*
+ * This is another ugly approach to keep the lock order, particularly
+ * mm->mmap_sem and aufs rwsem. The previous approach was reverted and you can
+ * find it in git-log, if you want.
+ *
+ * native readdir: i_mutex, copy_to_user, mmap_sem
+ * aufs readdir: i_mutex, rwsem, nested-i_mutex, copy_to_user, mmap_sem
+ *
+ * Before aufs_mmap() mmap_sem is acquired already, but aufs_mmap() has to
+ * acquire aufs rwsem. It introduces a circular locking dependency.
+ * To address this problem, aufs_mmap() delegates the part which requires aufs
+ * rwsem to its internal workqueue.
+ */
+
+/* very ugly approach */
+#include "mtx.h"
+
+struct au_mmap_pre_args {
+	/* input */
+	struct file *file;
+	struct vm_area_struct *vma;
+
+	/* output */
+	int *errp;
+	struct file *h_file;
+	struct au_branch *br;
+	int mmapped;
+};
+
+static int au_mmap_pre(struct file *file, struct vm_area_struct *vma,
+		       struct file **h_file, struct au_branch **br,
+		       int *mmapped)
 {
 	int err;
-	struct vm_operations_struct *h_ops;
-
-	AuRwMustAnyLock(&finfo->fi_rwsem);
-
-	err = 0;
-	h_ops = finfo->fi_h_vm_ops;
-	AuDebugOn(!h_ops);
-	if ((!h_ops->page_mkwrite && !h_ops->close)
-	    || finfo->fi_vm_ops)
-		goto out;
-
-	err = -ENOMEM;
-	finfo->fi_vm_ops = kmemdup(&aufs_vm_ops, sizeof(aufs_vm_ops), GFP_NOFS);
-	if (unlikely(!finfo->fi_vm_ops))
-		goto out;
-
-	err = 0;
-	if (h_ops->page_mkwrite)
-		finfo->fi_vm_ops->page_mkwrite = aufs_page_mkwrite;
-	if (h_ops->close)
-		finfo->fi_vm_ops->close = aufs_vm_close;
-
-	vma->vm_ops = finfo->fi_vm_ops;
-
- out:
-	return err;
-}
-
-static int aufs_mmap(struct file *file, struct vm_area_struct *vma)
-{
-	int err;
-	unsigned char wlock, mmapped;
+	aufs_bindex_t bstart;
+	const unsigned char wlock
+		= !!(file->f_mode & FMODE_WRITE) && (vma->vm_flags & VM_SHARED);
 	struct dentry *dentry;
 	struct super_block *sb;
-	struct file *h_file;
-	struct vm_operations_struct *vm_ops;
 
 	dentry = file->f_dentry;
-	wlock = !!(file->f_mode & FMODE_WRITE) && (vma->vm_flags & VM_SHARED);
 	sb = dentry->d_sb;
-	si_read_lock(sb, AuLock_FLUSH);
+	si_read_lock(sb, !AuLock_FLUSH);
 	err = au_reval_and_lock_fdi(file, au_reopen_nondir, /*wlock*/1);
 	if (unlikely(err))
 		goto out;
 
-	mmapped = !!au_test_mmapped(file);
+	*mmapped = !!au_test_mmapped(file);
 	if (wlock) {
 		struct au_pin pin;
 
 		err = au_ready_to_write(file, -1, &pin);
-		di_downgrade_lock(dentry, AuLock_IR);
+		di_write_unlock(dentry);
 		if (unlikely(err))
 			goto out_unlock;
 		au_unpin(&pin);
 	} else
-		di_downgrade_lock(dentry, AuLock_IR);
+		di_write_unlock(dentry);
+	bstart = au_fbstart(file);
+	*br = au_sbr(sb, bstart);
+	*h_file = au_hf_top(file);
+	get_file(*h_file);
+	au_fi_mmap_lock(file);
 
-	h_file = au_h_fptr(file, au_fbstart(file));
-	if (!mmapped && au_test_fs_bad_mapping(h_file->f_dentry->d_sb)) {
+out_unlock:
+	fi_write_unlock(file);
+out:
+	si_read_unlock(sb);
+	return err;
+}
+
+static void au_call_mmap_pre(void *args)
+{
+	struct au_mmap_pre_args *a = args;
+	*a->errp = au_mmap_pre(a->file, a->vma, &a->h_file, &a->br,
+			       &a->mmapped);
+}
+
+static int aufs_mmap(struct file *file, struct vm_area_struct *vma)
+{
+	int err, wkq_err;
+	unsigned long h_vmflags;
+	struct au_finfo *finfo;
+	struct dentry *h_dentry;
+	struct vm_operations_struct *h_vmop, *vmop;
+	struct au_mmap_pre_args args = {
+		.file		= file,
+		.vma		= vma,
+		.errp		= &err
+	};
+
+	wkq_err = au_wkq_wait_pre(au_call_mmap_pre, &args);
+	if (unlikely(wkq_err))
+		err = wkq_err;
+	if (unlikely(err))
+		goto out;
+	finfo = au_fi(file);
+	mutex_set_owner(&finfo->fi_mmap);
+
+	h_dentry = args.h_file->f_dentry;
+	if (!args.mmapped && au_test_fs_bad_mapping(h_dentry->d_sb)) {
 		/*
 		 * by this assignment, f_mapping will differs from aufs inode
 		 * i_mapping.
 		 * if someone else mixes the use of f_dentry->d_inode and
 		 * f_mapping->host, then a problem may arise.
 		 */
-		file->f_mapping = h_file->f_mapping;
+		file->f_mapping = args.h_file->f_mapping;
 	}
 
-	vm_ops = NULL;
-	if (!mmapped) {
-		vm_ops = au_vm_ops(h_file, vma);
-		err = PTR_ERR(vm_ops);
-		if (IS_ERR(vm_ops))
-			goto out_unlock;
-	}
+	/* always try this internal mmap to get vma flags */
+	h_vmflags = 0; /* gcc warning */
+	h_vmop = au_hvmop(args.h_file, vma, &h_vmflags);
+	err = PTR_ERR(h_vmop);
+	if (IS_ERR(h_vmop))
+		goto out_unlock;
+	AuDebugOn(args.mmapped && h_vmop != finfo->fi_hvmop);
+
+	vmop = (void *)au_dy_vmop(file, args.br, h_vmop);
+	err = PTR_ERR(vmop);
+	if (IS_ERR(vmop))
+		goto out_unlock;
 
 	/*
 	 * unnecessary to handle MAP_DENYWRITE and deny_write_access()?
@@ -601,26 +670,19 @@ static int aufs_mmap(struct file *file, struct vm_area_struct *vma)
 	if (unlikely(err))
 		goto out_unlock;
 
-	vma->vm_ops = &aufs_vm_ops;
-	if (!mmapped) {
-		struct au_finfo *finfo = au_fi(file);
+	vma->vm_ops = vmop;
+	vma->vm_flags = h_vmflags;
+	if (!args.mmapped)
+		finfo->fi_hvmop = h_vmop;
 
-		finfo->fi_h_vm_ops = vm_ops;
-		mutex_init(&finfo->fi_vm_mtx);
-	}
-
-	err = au_custom_vm_ops(au_fi(file), vma);
-	if (unlikely(err))
-		goto out_unlock;
-
-	vfsub_file_accessed(h_file);
-	fsstack_copy_attr_atime(dentry->d_inode, h_file->f_dentry->d_inode);
+	vfsub_file_accessed(args.h_file);
+	/* update without lock, I don't think it a problem */
+	fsstack_copy_attr_atime(file->f_dentry->d_inode, h_dentry->d_inode);
 
  out_unlock:
-	di_read_unlock(dentry, AuLock_IR);
-	fi_write_unlock(file);
+	au_fi_mmap_unlock(file);
+	fput(args.h_file);
  out:
-	si_read_unlock(sb);
 	return err;
 }
 
@@ -660,7 +722,7 @@ static int aufs_fsync_nondir(struct file *file, struct dentry *dentry,
 	au_unpin(&pin);
 
 	err = -EINVAL;
-	h_file = au_h_fptr(file, au_fbstart(file));
+	h_file = au_hf_top(file);
 	if (h_file->f_op && h_file->f_op->fsync) {
 		struct dentry *h_d;
 		struct mutex *h_mtx;
@@ -725,7 +787,7 @@ static int aufs_aio_fsync_nondir(struct kiocb *kio, int datasync)
 	au_unpin(&pin);
 
 	err = -ENOSYS;
-	h_file = au_h_fptr(file, au_fbstart(file));
+	h_file = au_hf_top(file);
 	if (h_file->f_op && h_file->f_op->aio_fsync) {
 		struct dentry *h_d;
 		struct mutex *h_mtx;
@@ -770,7 +832,7 @@ static int aufs_fasync(int fd, struct file *file, int flag)
 	if (unlikely(err))
 		goto out;
 
-	h_file = au_h_fptr(file, au_fbstart(file));
+	h_file = au_hf_top(file);
 	if (h_file->f_op && h_file->f_op->fasync)
 		err = h_file->f_op->fasync(fd, h_file, flag);
 
@@ -795,6 +857,7 @@ static ssize_t aufs_sendpage(struct file *file, struct page *page, int offset,
 /* ---------------------------------------------------------------------- */
 
 const struct file_operations aufs_file_fop = {
+	.owner		= THIS_MODULE,
 	/*
 	 * while generic_file_llseek/_unlocked() don't use BKL,
 	 * don't use it since it operates file->f_mapping->host.
@@ -808,10 +871,14 @@ const struct file_operations aufs_file_fop = {
 	.aio_write	= aufs_aio_write,
 #ifdef CONFIG_AUFS_POLL
 	.poll		= aufs_poll,
+#endif
+	.unlocked_ioctl	= aufs_ioctl_nondir,
+#ifdef CONFIG_COMPAT
+	.compat_ioctl	= aufs_ioctl_nondir, /* same */
 #endif
 	.mmap		= aufs_mmap,
 	.open		= aufs_open_nondir,
-	.flush		= aufs_flush,
+	.flush		= aufs_flush_nondir,
 	.release	= aufs_release_nondir,
 	.fsync		= aufs_fsync_nondir,
 	/* .aio_fsync	= aufs_aio_fsync_nondir, */
